import java.net.*;
import java.io.*;
import java.util.*;

/**
 * A lock manager to manage shared and exclusive locks. The lock
 * manager will queue the requests which are not granted. The lock
 * will be maintained in ascending order by txnId.
 *
 * @author  Jyrki Nummenmaa
 * @version 0.1
 * @see     LMMultiServerThread
 * @see     LockManagerProtocol
 */

public class LockManager extends Thread
{

	public static final int NO_TXN = -1;
	public static final int NO_ITEM = -1;
	public static final int LOCKED = 1;
	public static final int QUEUED = 2;
	public static final int ALREADY_LOCKED = 3;
	public static final int FAILED = -1;


	/* TS */
	public int serverId = -1;

	/** 0 = no output, 1 = output lock/unlockqueue events and client messages, 
	 *  2 = even more but seemingly randomly chosen output :) 
	 */
	private int debug = 1;

	/** If true, keep on listening to the server socket. Else exit. */
	private boolean listening = true;

	/** Lock objects for currently locked objects */
	private Vector locks = new Vector();
	/** Lock objects for locks in queue */
	private Vector lockQueue = new Vector();
	/** Lock objects for locks given from queue to be notified to the client */
	private Vector notifyLocks = new Vector();

	/** Argument vector passed on from the program to start this thread. */
	private String[] args;

	/** Just stores the argument String vector reference. */
	public LockManager(String[] args)
	{
		this.args = args;
	}

	/* TS */
	public Vector getQueue()
	{
		return lockQueue;
	}
	public Vector getLocks()
	{
		return locks;
	}


	/** Checks, if is a lock with itemNo and lockType and txn id != txnId */
	/* TS
	private synchronized boolean isLockedByOthers(int txnId, int itemNo, boolean lockType) */
	private synchronized boolean isLockedByOthers(long txnId, int itemNo, boolean lockType) 
	{
		for (int i = 0; i < locks.size(); i++)
		{
			Lock l = (Lock)locks.elementAt(i);
			if ((l.getItemNo() == itemNo)
				&& (l.getTxnId() != txnId)
				&& (l.getLockType() == lockType))
				return true;
		}
		return false;
	}

	/** Checks, if is a lock with itemNo and txn id != txnId */
	/* TS
	private synchronized boolean isLockedByOthers(int txnId, int itemNo) */
	private synchronized boolean isLockedByOthers(long txnId, int itemNo)
	{
		for (int i = 0; i < locks.size(); i++)
		{
			Lock l = (Lock)locks.elementAt(i);
			if ((l.getItemNo() == itemNo) && (l.getTxnId() != txnId))
				return true;
		}
		return false;
	}

	/** Checks, if there is a lock with txnId, itemNo, and lockType. */
	/* TS
	private synchronized boolean isLockedByTxn(int txnId, int itemNo, boolean lockType) */
	private synchronized boolean isLockedByTxn(long txnId, int itemNo, boolean lockType)
	{
		for (int i = 0; i < locks.size(); i++)
		{
			Lock l = (Lock)locks.elementAt(i);
			if ((l.getItemNo() == itemNo)
				&& (l.getTxnId() == txnId)
				&& (l.getLockType() == lockType))
				return true;
		}
		return false;
	}

	/** Checks, if there is a lock with txnId, itemNo. */
	/* TS
	private synchronized boolean isLockedByTxn(int txnId, int itemNo) */
	private synchronized boolean isLockedByTxn(long txnId, int itemNo)
	{
		for (int i = 0; i < locks.size(); i++)
		{
			Lock l = (Lock)locks.elementAt(i);
			if ((l.getItemNo() == itemNo) && (l.getTxnId() == txnId))
				return true;
		}
		return false;
	}

	/** Checks, if oldLock is in lockQueue. */
	private synchronized boolean isInQueue(Lock oldLock)
	{
		for (int i = 0; i < locks.size(); i++)
		{
			Lock l = (Lock)locks.elementAt(i);
			if (l.equals(oldLock))
				return true;
		}
		return false;
	}
	
/* TS */
	public synchronized boolean isTxnInQueue(long txnId)
	{
		Lock lc;
		for(int i = 0; i < lockQueue.size(); i++)
		{
			lc = (Lock)lockQueue.elementAt(i);
			if(lc.getTxnId() == txnId)
				return true;
		}
		return false;
	}

	/** Adds newLock into lockQueue keeping lockQueue in ascending order by txnId. */
	private synchronized void addToQueue(Lock newLock)
	{
		boolean found = false;
		int i = 0;
		while ((!found) && (i < lockQueue.size()))
		{
			Lock l = (Lock)lockQueue.elementAt(i);
			if (l.getTxnId() > newLock.getTxnId())
			{ //insert before this to keep txn order!
				lockQueue.add(i, newLock);
				found = true;
			}
			i++;
		}
		if (!found)
			lockQueue.add(i, newLock);
		if (debug > 0)
			System.out.println(
				"- Added "
					+ newLock.outputString()
					+ " to lock queue at position "
					+ i
					+ ".");
	}

	/** Removes oldLock from lockQueue. */
	private synchronized boolean removeFromQueue(Lock oldLock)
	{
		for (int i = 0; i < lockQueue.size(); i++)
		{
			Lock l = (Lock)lockQueue.elementAt(i);
			if (l.equals(oldLock))
			{
				lockQueue.remove(i);
				if (debug > 0)
					System.out.println(
						"- Removed "
							+ oldLock.outputString()
							+ " from lock queue.");
				return true;
			}
		}
		return false;
	}

	/** 
	 * Returns itemNo for one lock with id txnId and removes that lock from notifyLocks,
	 * if such a lock is found. Otherwise, returns NO_ITEM. 
	 */
	/* TS
	public synchronized int notifyLock(int txnId, boolean lockType) */
	public synchronized int notifyLock(long txnId, boolean lockType)
	{
		// I just return an int to avoid references to volatile data -
		// if I return a reference and the lock is gone before the actual
		// notification takes place (unlikely, though), then it is bad :(
		int i = 0;
		while (i < notifyLocks.size())
		{
			Lock l = (Lock)notifyLocks.elementAt(i);
			if ((l.getTxnId() == txnId) && (l.getLockType() == lockType))
			{
				if (debug > 1)
					System.out.println(
						"- Found a lock to be notified: " + l.outputString());
				notifyLocks.remove(i);
				return l.getItemNo();
			}
			i++;
		}
		return NO_ITEM;
	}

	/** Changes all possible lockQueue items into real locks (locks items). */
	private synchronized void lockFromQueue(int itemNo)
	{
		if (debug > 1)
			System.out.println(
				"- Checking lock queue for requests on item " + itemNo);
		int txnId;
		boolean lockType;
		int i = 0;
		while (i < lockQueue.size())
		{
			boolean found = false;
			Lock l = (Lock)lockQueue.elementAt(i);
			if (debug > 1)
				System.out.println(
					"- Checking lock queue item : " + l.outputString());
			if (l.getItemNo() == itemNo)
				if (l.getLockType() == Lock.SHARED)
				{
					if (!isLockedByOthers(l.getTxnId(),
						itemNo,
						Lock.EXCLUSIVE))
					{
						if (debug > 0)
							System.out.println(
								"- Locking "
									+ l.outputString()
									+ " from lock queue.");
						found = true;
						locks.add(l);
						notifyLocks.add(l);
						lockQueue.remove(i);
					}
				}
				else // EXCLUSIVE
					if (!isLockedByOthers(l.getTxnId(), itemNo))
					{
						if (debug > 0)
							System.out.println(
								"- Locking "
									+ l.outputString()
									+ " from lock queue.");
						found = true;
						locks.add(l);
						notifyLocks.add(l);
						lockQueue.remove(i);
					}
			if (!found)
				i++;
		}
	}

	/** Returns the current debug value. */
	public int getDebug()
	{
		return debug;
	}

	/** 
	 * Processes the lock request on Lock newLock. If the txn already has the lock,
	 * returns ALREADY_LOCKED. If the txn does not have the lock, and there is no
	 * lock conflict, adds the lock and returns LOCKED. If the txn does not have the
	 * lock, puts the lock into lockQueue and returns QUEUED. Already having a SHARED
	 * lock and requesting an EXCLUSIVE lock is the same as not having the lock before.
	 */
	public synchronized int lock(Lock newLock)
	{
		if (debug > 1)
			System.out.println(
				"- Lock requested for " + newLock.outputString());
		if (newLock.getLockType() == Lock.SHARED)
		{
			if (isLockedByTxn(newLock.getTxnId(),
				newLock.getItemNo(),
				Lock.SHARED)
				|| isLockedByTxn(
					newLock.getTxnId(),
					newLock.getItemNo(),
					Lock.EXCLUSIVE))
			{
				if (debug > 0)
					System.out.println(
						"- Item was already locked: " + newLock.outputString());
				return ALREADY_LOCKED; // was already locked by txnId
			}
			if (!isLockedByOthers(newLock.getTxnId(),
				newLock.getItemNo(),
				Lock.EXCLUSIVE))
			{
				locks.add(newLock);
				if (debug > 0)
					System.out.println(
						"- Adding a new lock " + newLock.outputString());
				return LOCKED;
			}
			else
			{
				if (debug > 0)
					System.out.println(
						"- Queueing a new lock " + newLock.outputString());
				addToQueue(newLock);
				return QUEUED;
			}
		}
		else
		{ // EXCLUSIVE
			if (isLockedByTxn(newLock.getTxnId(),
				newLock.getItemNo(),
				Lock.EXCLUSIVE))
			{
				if (debug > 0)
					System.out.println(
						"- Item was already locked: " + newLock.outputString());
				return ALREADY_LOCKED; // was already locked by txnId
			}
			if (!isLockedByOthers(newLock.getTxnId(), newLock.getItemNo()))
			{
				locks.add(newLock);
				if (debug > 0)
					System.out.println(
						"- Adding a new lock " + newLock.outputString());
				return LOCKED;
			}
			else
			{
				if (debug > 0)
					System.out.println(
						"- Queueing a new lock " + newLock.outputString());
				addToQueue(newLock);
				return QUEUED;
			}
		}
	}

	/** 
	 * Removes all locks with txnId and itemNo and removes all queued locks with
	 * txnId and itemNo. After this, calls lockFromQueue(itemNo).
	 */
	/* TS
	public synchronized boolean unlock(int txnId, int itemNo) */
	public synchronized boolean unlock(long txnId, int itemNo)
	{
		if (debug > 1)
			System.out.println(
				"- Unlock requested for item " + itemNo + " by txn " + txnId);
		if (!isLockedByTxn(txnId, itemNo))
			return false;
		int i = 0;
		while (i < locks.size())
		{ // removing all locks for with txnId, itemNo
			Lock l = (Lock)locks.elementAt(i);
			if ((l.getItemNo() == itemNo) && (l.getTxnId() == txnId))
			{
				if (debug > 0)
					System.out.println("- Removing lock: " + l.outputString());
				locks.remove(i);
			}
			i++;
		}
		i = 0;
		while (i < lockQueue.size())
		{ // removing all lockQueue items with txnId, itemNo
			Lock l = (Lock)lockQueue.elementAt(i);
			if ((l.getItemNo() == itemNo) && (l.getTxnId() == txnId))
			{
				if (debug > 0)
					System.out.println(
						"- Removing queued lock: " + l.outputString());
				lockQueue.remove(i);
			}
			i++;
		}
		if (debug > 1)
			System.out.println("- Locks removed.");
		lockFromQueue(itemNo);
		if (debug > 1)
			System.out.println("- Queues checked.");
		return true;
	}

	/* TS */
	public synchronized long getTxnId()
	{
		long txnId = System.currentTimeMillis(); // Get sytem time
		txnId = txnId << 16; // Shift left by 16 bits
		txnId = txnId | serverId; // Add server id into the txnId
		return txnId;
	}


	/** 
	 * Starts listening for a ServerSocket and creates a new LMMultiServerThread 
	 * for each connection.
	 */

	public void run()
	{
		ServerSocket serverSocket = null;
		int portNo = 0;

		try
		{
			portNo = Integer.parseInt(args[0]);
			serverSocket = new ServerSocket(portNo);
			/* TS */
			serverId = Integer.parseInt(args[1]);
			/* TS
			if (args.length > 1) */
			if (args.length > 2)
				debug = Integer.parseInt(args[2]);
			while (listening)
				new LMMultiServerThread(this, serverSocket.accept()).start();
			serverSocket.close();

		}
		catch (IOException e)
		{
			System.err.println("Could not listen on port:" + portNo);
			System.exit(-1);
		}
		catch (Exception e)
		{
			System.err.println("Could not listen on port:" + portNo);
			System.exit(-1);
		}
	}

}
