import java.util.*;

/**
 * A class to manage the communication protocol with a lock client.
 *
 * @author  Jyrki Nummenmaa
 * @version 0.1
 * @see     LockManager
 * @see     LockManagerProtocol
 */

public class LockManagerProtocol
{

	private static final int NO_TXN = -1;
	private static final int NO_ITEM = -1;

	/** Shared lock command */
	private final String sLockCommand = "slock";
	/** Exclusive lock command */
	private final String xLockCommand = "xlock";
	/** Unlock command */
	private final String unlockCommand = "unlock";
	/** Command for closing connection */
	private final String byeCommand = "bye";
	/** Command for communicating txn id */
	private final String txnCommand = "txn";

	/* TS 
	private int txnId = NO_TXN; */
	private long txnId = NO_TXN;

	private LockManager lm;

	public LockManagerProtocol(LockManager lm)
	{
		this.lm = lm;
	}

	/** 
	 * Pulls a single lock notification from the notify queue
	 * and creates the response string for client.
	 */
	public String checkNotify()
	{
		int itemNo = lm.notifyLock(txnId, Lock.EXCLUSIVE);
		if (itemNo != LockManager.NO_ITEM)
			return "xlocked " + itemNo + " - gave queued exclusive lock";
		itemNo = lm.notifyLock(txnId, Lock.SHARED);
		if (itemNo != LockManager.NO_ITEM)
			return "slocked " + itemNo + " - gave queued shared lock";
		return null;
	}

	/** 
	 * Processes an input string from the client. Calls the respective
	 * LockManager's lock management methods and communicates the
	 * result to the client.
	 */
	public String processInput(String theInput)
	{
		String theOutput = null;

		if (theInput == null)
		{
			return "Connection established";
		}

		if (lm.getDebug() > 0)
			if (txnId != NO_TXN)
				System.out.println("Txn " + txnId + " says: " + theInput);
			else
				System.out.println("Got input: " + theInput);

		StringTokenizer st = new StringTokenizer(theInput);

		String command = "";
		int itemNo = NO_ITEM;

		try
		{
			if (st.hasMoreTokens())
				command = st.nextToken();
			if (command.equals(byeCommand))
				theOutput = "Bye";
			/* TS Commands for requesting TxnId and making 2PC */
			else if(command.equals("gettxnid"))
			{
				System.out.println("Getting txn id.. serverID: " + lm.serverId);				
								theOutput = "" + lm.getTxnId();
				System.out.println("TxnId: " + theOutput);			
			}
			else if(command.equals("prepare"))
			{
				theOutput = "yes";
				if(lm.isTxnInQueue(txnId))
					theOutput = "no";		
				System.out.println("Voted '" + theOutput + "' for " + 
					txnId + " (in queue " + lm.isTxnInQueue(txnId) + ")");
			}
			else if(command.equals("commit"))
			{
				System.out.println("COMMIT " + txnId);
				theOutput = "committed";
			}
			else if(command.equals("abort"))
			{
				System.out.println("ABORT " + txnId);
				theOutput = "aborted";
			}
			/* TS end */
			else if (command.equals(txnCommand))
			{
				txnId = Long.parseLong(st.nextToken());
				theOutput = "txn id " + txnId + " confirmed.";
			}
			else if (command.equals(sLockCommand))
			{
				itemNo = Integer.parseInt(st.nextToken());
				Lock l = new Lock(txnId, itemNo, Lock.SHARED);
				int success = lm.lock(l);
				if (success == LockManager.LOCKED)
					theOutput = "slocked " + itemNo;
				else if (success == LockManager.QUEUED)
					theOutput = "queued " + itemNo + " for a shared lock.";
				else // btw, this should never happen:
					theOutput = "denied lock for item " + itemNo;
			}
			else if (command.equals(xLockCommand))
			{
				itemNo = Integer.parseInt(st.nextToken());
				Lock l = new Lock(txnId, itemNo, Lock.EXCLUSIVE);
				int success = lm.lock(l);
				if (success == LockManager.LOCKED)
					theOutput = "xlocked " + itemNo;
				else if (success == LockManager.QUEUED)
					theOutput = "queued " + itemNo + " for a shared lock.";
				else // this should also never happen
					theOutput = "denied lock for item " + itemNo;
			}
			else if (command.equals(unlockCommand))
			{
				itemNo = Integer.parseInt(st.nextToken());
				boolean success = lm.unlock(txnId, itemNo);
				if (success)
					theOutput = "unlocked " + itemNo;
				else
					theOutput =
						"unlocking "
							+ itemNo
							+ " failed - you did not have the lock";
			}
			else
				theOutput = "unknown command: " + command;

		}
		catch (Exception e)
		{
			theOutput = "error in performing command: " + command;
			;
			System.out.println(e);
		}

		return theOutput;
	}
}
