import java.util.Timer;
import java.util.TimerTask;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * A client, which repeatedly sleeps for a delay (given as a parameter).
 * and then picks up an item number from [min,max], both given as parameter.
 * If it has requested a lock on that item, which it has not canceled, it
 * send a cancel message for that item. Else it sends a lock message for
 * that item. It uses a parameter to randomize, if it requests a shared
 * or a random lock. Randomisation seed is picked from the system clock.
 *
 * Usage:
 * java RandomLockClient txnId minItem maxItem lockingRounds xLockPercentage
 *  noXlockServers noSLockServers requestDelayMilliSecs server0 port0 ... serverN portN
 *
 * @author  Jyrki Nummenmaa
 * @version 0.1
 * @see     LockManager
 * @see     LockManagerProtocol
 */

public class RandomLockClient {
    
    int portNo;
    int txnId = Integer.MAX_VALUE;
    /** Min itemNo for requests */
    int minItem = 0; 
    /** Max itemNo for requests */
    int maxItem = 0; 
    /** Probability of a an xlock as a percentage */
    int xLockPercentage;
    /** How many servers to consult for an xlock */ 
    int noXLockServers; 
    /** How many servers to consult for an slock */
    int noSLockServers; 
    /** Sleep time between activities */
    int requestDelayMilliSecs; 
    /** How many locking/unlocking rounds to perform */
    int lockingRounds; 
    int noConnections;
    Socket[] lcSockets;
    PrintWriter[] lcOuts;
    BufferedReader[] lcIns;
    String[] mainArgs;
    boolean[] requests;
    boolean[][] requestServers;
    int noItems;
    Random random;

    /** Parse args, create connections and a Timer-based task for lock/unlock */
    public RandomLockClient(String[] args) {
	noConnections = (args.length-7)/2;
	lcSockets = new Socket[noConnections];
	lcOuts = new PrintWriter[noConnections];
	lcIns = new BufferedReader[noConnections];
	mainArgs = args;

        try {
	    txnId = Integer.parseInt(args[0]); 
	    minItem = Integer.parseInt(args[1]);
	    maxItem = Integer.parseInt(args[2]);
	    lockingRounds =  Integer.parseInt(args[3]);
	    xLockPercentage = Integer.parseInt(args[4]);
	    noXLockServers = Integer.parseInt(args[5]);
	    noSLockServers = Integer.parseInt(args[6]);
	    requestDelayMilliSecs = Integer.parseInt(args[7]);

	    noItems = maxItem-minItem+1;
	    
	    requests = new boolean[noItems]; // actice lock requests on item
	    for (int i=0;i<noItems;i++)
		requests[i] = false; 
	    
	    requestServers = new boolean[noItems][noConnections];
	    for (int i=0;i<noItems;i++)  
		for (int j=0;j<noConnections;j++)
		    requestServers[i][j]=false;
	    
	    
	    random = new Random(System.currentTimeMillis());
	    
	    
	    for (int i=0; i<noConnections; i++) {
		portNo = Integer.parseInt(mainArgs[2*i+9]);
		lcSockets[i] = new Socket(mainArgs[2*i+8], portNo);
		lcOuts[i] = new PrintWriter(lcSockets[i].getOutputStream(), true);
		lcIns[i] = new BufferedReader(new InputStreamReader(lcSockets[i].getInputStream()));
	    }

	    for (int i=0;i<noConnections;i++) {
		lcOuts[i].println("txn " + txnId);
	    }
	    
	} catch (UnknownHostException e) {
	    System.err.println("Don't know about host.");
	    System.exit(1);
	} catch (IOException e) {
	    System.err.println("Couldn't get I/O for the connection to the host.");
	    System.exit(1);
	} catch (Exception e) {
	    System.err.println(e);
	    System.exit(1);
	}
	

	Timer timer = new Timer();
	timer.schedule(new RandomLockTask(),
		       0,        //initial delay
	               1*requestDelayMilliSecs);  //subsequent rate
    }

    /** A task to do the lock/unlock activities */
    class RandomLockTask extends TimerTask {

	public void run() {
	    
	    if (lockingRounds>0) {
	    	
		String fromServer = "";
		String fromUser="";
		
		try {
		    
		    for (int i=0;i<noConnections;i++) {
			while (lcIns[i].ready()) {
			    fromServer = lcIns[i].readLine();
			    System.out.println("Server " + i + ": " + fromServer);
			}
		    }
		    
		    int itemNo = random.nextInt(noItems);
		    if (requests[itemNo]) { // there is a request, cancel it
			String unlockRequest = "unlock " + itemNo;
			for (int j=0;j<noConnections;j++) {
			    if (requestServers[itemNo][j]) {
				lcOuts[j].println(unlockRequest);
				requestServers[itemNo][j]=false;
				System.out.println("To server " + j + 
						   ":" + unlockRequest);
			    }
			}
		    }
		    else { // we have not requested the lock, so we will do so now
			requests[itemNo]=true;
			if ((random.nextInt(100)+1) <= xLockPercentage) { // then do an xlock
			    String xlockRequest = "xlock " + itemNo;
			    for (int done=0;done<noXLockServers;done++)
				while (true) { // do one new server and then break out
				    int nextServer = random.nextInt(noConnections);
				    if (requestServers[itemNo][nextServer]==false) {
					requestServers[itemNo][nextServer]=true;
					lcOuts[nextServer].println(xlockRequest);
					System.out.println("To server " + nextServer + 
							   ":" + xlockRequest);
					break;
				    }
				}
			}
			else { // randomisation advices us to do an slock
			    String slockRequest = "slock " + itemNo;
			    for (int done=0;done<noSLockServers;done++)
				while (true) { // do one new server and then break out
				    int nextServer = random.nextInt(noConnections);
				    if (requestServers[itemNo][nextServer]==false) {
					requestServers[itemNo][nextServer]=true;
					lcOuts[nextServer].println(slockRequest);
					System.out.println("To server " + nextServer + 
							   ":" + slockRequest);
					break;
				    }
				}
			}    
		    }
		    
		}
		catch (Exception e) {
		    System.out.println(e);
		}
		lockingRounds--;
	    }  // Else the locking rounds are done - remove locks and close input and output

	    else {
		System.out.println("Final cleanup");
		try {
		    for (int i=0;i<noItems;i++)  
			for (int j=0;j<noConnections;j++) {
			    if (requestServers[i][j])
				lcOuts[j].println("unlock " + i);
			}
		
		
		    for (int i=0;i<noConnections;i++) {
			lcIns[i].close();
			lcOuts[i].close();
			lcSockets[i].close();
		    }
		} catch (Exception e) {}
		System.out.println("Exiting...");
		System.exit(0);
	    }
	}
    }

    public static void main(String[] args) throws IOException {
	RandomLockClient rlc = new RandomLockClient(args);
    }
	
}

